@@ -5,6 +5,16 @@
 #include "test_precomp.hpp"
 namespace opencv_test { namespace {
 
+#if !defined CV_CXX11
+// Wrapper for generating seeded random number via std::rand.
+template<unsigned Seed>
+class SeededRandFunctor {
+public:
+    SeededRandFunctor() { std::srand(Seed); }
+    int operator()(int i) { return std::rand() % (i + 1); }
+};
+#endif
+
 std::string encode_qrcode_images_name[] = {
         "version1_mode1.png", "version1_mode2.png", "version1_mode4.png",
         "version2_mode1.png", "version2_mode2.png", "version2_mode4.png",
@@ -295,15 +305,15 @@ TEST(Objdetect_QRCode_Encode_Decode, regression)
                     corners[k].y = corners[k].y * height_ratio;
                 }
 
-                std::string output_info = "";
-                Mat straight_barcode;
 #ifdef HAVE_QUIRC
-                bool success = decodeQRCode(resized_src, corners, output_info, straight_barcode);
-                EXPECT_TRUE(success) << "The generated QRcode cannot be decoded." << " Mode: " << (int)mode<<
-                                        " version: " << version << " error correction level: " << (int)level;
-#endif
+                Mat straight_barcode;
+                std::string output_info = QRCodeDetector().decode(resized_src, corners, straight_barcode);
+                EXPECT_FALSE(output_info.empty())
+                    << "The generated QRcode cannot be decoded." << " Mode: " << (int)mode
+                    << " version: " << version << " error correction level: " << (int)level;
                 EXPECT_EQ(input_info, output_info) << "The generated QRcode is not same as test data." << " Mode: " << (int)mode <<
                                                         " version: " << version << " error correction level: " << (int)level;
+#endif
             }
         }
     }
@@ -318,9 +328,9 @@ TEST(Objdetect_QRCode_Encode_Kanji, regression)
     Mat qrcode;
 
     const int testing_versions = 3;
-    std::string input_infos[testing_versions] = {"\x82\xb1\x82\xf1\x82\xc9\x82\xbf\x82\xcd\x90\xa2\x8a\x45", // こんにちは世界
-                                                 "\x82\xa8\x95\xa0\x82\xaa\x8b\xf3\x82\xa2\x82\xc4\x82\xa2\x82\xdc\x82\xb7", // お腹が空いています
-                                                 "\x82\xb1\x82\xf1\x82\xc9\x82\xbf\x82\xcd\x81\x41\x8e\x84\x82\xcd\x8f\xad\x82\xb5\x93\xfa\x96\x7b\x8c\xea\x82\xf0\x98\x62\x82\xb5\x82\xdc\x82\xb7" // こんにちは、私は少し日本語を話します
+    std::string input_infos[testing_versions] = {"\x82\xb1\x82\xf1\x82\xc9\x82\xbf\x82\xcd\x90\xa2\x8a\x45", // "Hello World" in Japanese
+                                                 "\x82\xa8\x95\xa0\x82\xaa\x8b\xf3\x82\xa2\x82\xc4\x82\xa2\x82\xdc\x82\xb7", // "I am hungry" in Japanese
+                                                 "\x82\xb1\x82\xf1\x82\xc9\x82\xbf\x82\xcd\x81\x41\x8e\x84\x82\xcd\x8f\xad\x82\xb5\x93\xfa\x96\x7b\x8c\xea\x82\xf0\x98\x62\x82\xb5\x82\xdc\x82\xb7" // "Hello, I speak a little Japanese" in Japanese
                                                 };
 
     for (int i = 0; i < testing_versions; i++)
@@ -345,13 +355,12 @@ TEST(Objdetect_QRCode_Encode_Kanji, regression)
             corners[j].y = corners[j].y * height_ratio;
         }
 
-        Mat straight_barcode;
-        std::string decoded_info = "";
-
 #ifdef HAVE_QUIRC
-        EXPECT_TRUE(decodeQRCode(resized_src, corners, decoded_info, straight_barcode)) << "The generated QRcode cannot be decoded.";
-#endif
+        Mat straight_barcode;
+        std::string decoded_info = QRCodeDetector().decode(resized_src, corners, straight_barcode);
+        EXPECT_FALSE(decoded_info.empty()) << "The generated QRcode cannot be decoded.";
         EXPECT_EQ(input_info, decoded_info);
+#endif
     }
 }
 
@@ -381,8 +390,15 @@ TEST(Objdetect_QRCode_Encode_Decode_Structured_Append, DISABLED_regression)
         std::string symbol_set = config["symbols_set"];
 
         std::string input_info = symbol_set;
-        std::random_shuffle(input_info.begin(), input_info.end());
-
+#if defined CV_CXX11
+        // std::random_shuffle is deprecated since C++11 and removed in C++17.
+        // Use manually constructed RNG with a fixed seed and std::shuffle instead.
+        std::mt19937 rand_gen {1};
+        std::shuffle(input_info.begin(), input_info.end(), rand_gen);
+#else
+        SeededRandFunctor<1> rand_gen;
+        std::random_shuffle(input_info.begin(), input_info.end(), rand_gen);
+#endif
         for (int j = min_stuctures_num; j < max_stuctures_num; j++)
         {
             QRCodeEncoder::Params params;
@@ -413,17 +429,21 @@ TEST(Objdetect_QRCode_Encode_Decode_Structured_Append, DISABLED_regression)
                     corners[m].y = corners[m].y * height_ratio;
                 }
 
-                std::string decoded_info = "";
-                Mat straight_barcode;
 #ifdef HAVE_QUIRC
-                bool success = decodeQRCode(resized_src, corners, decoded_info, straight_barcode);
-                EXPECT_TRUE(success) << "The generated QRcode cannot be decoded." << " Mode: " << mode <<
-                                        " structures number: " << k << "/" << j;
-#endif
+                Mat straight_barcode;
+                std::string decoded_info = QRCodeDetector().decode(resized_src, corners, straight_barcode);
+                EXPECT_FALSE(decoded_info.empty())
+                    << "The generated QRcode cannot be decoded." << " Mode: " << modes[i]
+                    << " structures number: " << k << "/" << j;
                 output_info += decoded_info;
+#endif
             }
+#ifdef HAVE_QUIRC
             EXPECT_EQ(input_info, output_info) << "The generated QRcode is not same as test data." << " Mode: " << mode <<
                                                   " structures number: " << j;
+#else
+            std::cout << "Mode=" << mode << ": Unable to verify generated QR codes - QUIRC is disabled" << std::endl;
+#endif
         }
     }
 }
